// 4-Digit Counter
// 2.9.2012
// metal @ diyaudio.com
// compiler: AVR-GCC 4.7.2

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h> 

#define SEGMENT_PORT	PORTB
#define SEGMENT_PIN		PINB
#define SEGMENT_DDR		DDRB

#define DIGIT_PORT		PORTD
#define DIGIT_PIN		PIND
#define	DIGIT_DDR		DDRD	

// refresh frequncy
#define	REFRESH_FREQ	250

#define KEY_PORT        PORTD
#define KEY_PIN         PIND
#define	KEY_DDR			DDRD
#define KEY_MAIN		6
#define KEY_PLUS		5
#define KEY_MINUS		4
#define ALL_KEYS        (1<<KEY_MAIN | 1<<KEY_PLUS | 1<<KEY_MINUS)

#define REPEAT_MASK     (1<<KEY_MAIN)       // repeat function
#define REPEAT_START    150                	// after 500ms
#define REPEAT_NEXT     10                	// every 200ms

// Timer0 prescaler values
#define	TMR0_OFF		~((1<<CS02) | (1<<CS01) | (1<<CS00))
#define NO_PRESCALE		(1<<CS00)
#define	PRESCALE_8		(1<<CS01)
#define	PRESCALE_64		(1<<CS01) | (1<<CS00)
#define	PRESCALE_256	(1<<CS02)
#define	PRESCALE_1024	(1<<CS02) | (1<<CS00)	

// display data
#define SEG_a	0x01
#define SEG_b	0x04
#define SEG_c	0x10
#define SEG_d	0x40
#define SEG_e	0x80
#define SEG_f	0x02
#define SEG_g	0x08
#define SEG_dot	0x20

// common anode or common cathode
// 1 --> common anode
// 0 --> common cathode
#define CA	1

volatile unsigned int val = 0;

#if CA
// & ~SEG_dot
unsigned char digits[] = {
	~(SEG_a|SEG_b|SEG_c|SEG_d|SEG_e|SEG_f),			// 0
	~(SEG_b|SEG_c),									// 1
	~(SEG_a|SEG_b|SEG_d|SEG_e|SEG_g),				// 2
	~(SEG_a|SEG_b|SEG_c|SEG_d|SEG_g),				// 3
	~(SEG_b|SEG_c|SEG_c|SEG_f|SEG_g),				// 4
	~(SEG_a|SEG_c|SEG_d|SEG_f|SEG_g),				// 5
	~(SEG_a|SEG_c|SEG_d|SEG_e|SEG_f|SEG_g),			// 6
	~(SEG_a|SEG_b|SEG_c),							// 7
	~(SEG_a|SEG_b|SEG_c|SEG_d|SEG_e|SEG_f|SEG_g),	// 8
	~(SEG_a|SEG_b|SEG_c|SEG_d|SEG_f|SEG_g),			// 9
};
#else
// | SEG_dot
unsigned char digits[] = {
	(SEG_a|SEG_b|SEG_c|SEG_d|SEG_e|SEG_f),			// 0
	(SEG_b|SEG_c),									// 1
	(SEG_a|SEG_b|SEG_d|SEG_e|SEG_g),				// 2
	(SEG_a|SEG_b|SEG_c|SEG_d|SEG_g),				// 3
	(SEG_b|SEG_c|SEG_c|SEG_f|SEG_g),				// 4
	(SEG_a|SEG_c|SEG_d|SEG_f|SEG_g),				// 5
	(SEG_a|SEG_c|SEG_d|SEG_e|SEG_f|SEG_g),			// 6
	(SEG_a|SEG_b|SEG_c),							// 7
	(SEG_a|SEG_b|SEG_c|SEG_d|SEG_e|SEG_f|SEG_g),	// 8
	(SEG_a|SEG_b|SEG_c|SEG_d|SEG_f|SEG_g),			// 9
};
#endif

volatile unsigned char 	key_state;              // debounced and inverted key state:
												// bit = 1: key pressed
volatile unsigned char 	key_press;              // key press detect
volatile unsigned char 	key_rpt;                // key long press and repeat

//#define ACTIVE_HIGH		0

ISR(TIMER0_COMPA_vect) 
{
	unsigned char cnt = 0;
	static unsigned char active_digit = 0; 
	unsigned char digit_val[4] = {0, 0, 0, 0}; // NUMBER OF DIGITS
	unsigned int tmp;
	// static unsigned char blink = 0;
	
	// turn all digits off
	#if CA
	DIGIT_PORT |= 0x0F;	
	#else
	DIGIT_PORT &= 0xF0;	
	#endif
	
	if ( val == 0) cnt = 1;
	
	// extract each digit's value
	tmp = val;	// necessary to prevent modifying val variable
     while (tmp > 0)
	{
		digit_val[cnt++] = tmp % 10;
		tmp /= 10;
    }	


	// suppress leading zeros
	active_digit = ++active_digit % cnt;
	SEGMENT_PORT = digits[digit_val[ active_digit]];
	// bliking a specified digit
	//if (!(active_digit == 0 && blink > 124))

	// turn corresponding digit on
	#if CA
	DIGIT_PORT &= ~(0x01 << active_digit);
	#else
	DIGIT_PORT |= (0x01 << active_digit);
	#endif
	//blink += 2;
	
	// *** debounce code ***
  static unsigned char ct0, ct1;
  static unsigned int rpt;
  unsigned char i;

#ifdef ACTIVE_HIGH
  // key changed ?
  i = key_state ^ KEY_PIN;
#else
  // key changed ?
  i = key_state ^ ~KEY_PIN;
#endif
  // reset or count ct0
  ct0 = ~( ct0 & i );
  // reset or count ct1
  ct1 = ct0 ^ (ct1 & i);
  // count until roll over ?
  i &= ct0 & ct1;
  // then toggle debounced state
  key_state ^= i;
  // 0->1: key press detect
  key_press |= key_state & i;
  // check repeat function
  if ((key_state & REPEAT_MASK) == 0)
  // start delay
  rpt = REPEAT_START;
  if (--rpt == 0)
  {
    // repeat delay
    rpt = REPEAT_NEXT;
#ifdef ACTIVE_HIGH
    key_rpt = ~(key_rpt | (key_state & REPEAT_MASK));
#else
    key_rpt |= key_state & REPEAT_MASK;
#endif
  }
}

///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed. Each pressed key is reported
// only once
//
unsigned char get_key_press(unsigned char key_mask)
{
  cli();                                          // read and clear atomic !
  key_mask &= key_press;                          // read key(s)
  key_press ^= key_mask;                          // clear key(s)
  sei();
  return key_mask;
}
 
///////////////////////////////////////////////////////////////////
//
// check if a key has been pressed long enough such that the
// key repeat functionality kicks in. After a small setup delay
// the key is reported beeing pressed in subsequent calls
// to this function. This simulates the user repeadiately
// pressing and releasing the key.
//
unsigned char get_key_rpt(unsigned char key_mask)
{
  cli();                                          // read and clear atomic !
  key_mask &= key_rpt;                            // read key(s)
  key_rpt ^= key_mask;                            // clear key(s)
  sei();
  return key_mask;
}
 
///////////////////////////////////////////////////////////////////
//
unsigned char get_key_short(unsigned char key_mask)
{
  cli();                                          // read key state and key press atomic !
  return get_key_press(~key_state & key_mask);
}
 
///////////////////////////////////////////////////////////////////
//
unsigned char get_key_long(unsigned char key_mask)
{
  return get_key_press(get_key_rpt(key_mask));
}

int main()
{
	SEGMENT_DDR = 0xFF;
	DIGIT_DDR = 0x0F;
	
	//Configure key port for input
	KEY_DDR &= ~ALL_KEYS;
	// and turn on pull up resistors
	KEY_PORT |= ALL_KEYS;
		
	// turn all digits off
	#if CA
	DIGIT_PORT |= 0x0F;	
	#else
	DIGIT_PORT &= 0xF0;	
	#endif
	
	

	TIMSK |= (1 << OCIE0A); // enable Timer0 output compare match interrupt 
	// calculate compare value
	// 256 is the prescaler value
	OCR0A = F_CPU / 256 / REFRESH_FREQ;
	// CTC mode
	TCCR0A |= (1<<WGM01);
	//prescaler = 256
	TCCR0B |= PRESCALE_256;
	sei(); // enable global interrupts 
	val = 0;
	
	while(1)
	{
		if(get_key_press( 1<<KEY_MAIN ))
		{
			val=0;
		}
		if(get_key_press( 1<<KEY_PLUS ))
		{
			val++;
		}
		if(get_key_press( 1<<KEY_MINUS ))
		{
			val--;
		}
		if (val > 9999) val = 0;
	}
}
